# Copyright (c) 2025 syncblaze
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""A simple rate limiter for FastAPI."""



from math import ceil

from fastapi import HTTPException
from starlette.requests import Request
from starlette.responses import Response
from starlette.status import HTTP_429_TOO_MANY_REQUESTS
from starlette.websockets import WebSocket

from act_ratelimit.backends import BaseBackend
from act_ratelimit.constants import HTTP_CALLBACK_SIGNATURE
from act_ratelimit.constants import IDENTIFIER_SIGNATURE
from act_ratelimit.constants import WS_CALLBACK_SIGNATURE
from act_ratelimit.constants import RateLimitStrategy

__all__: tuple[str, ...] = ("ACTRatelimit",)


async def default_identifier(request: Request | WebSocket) -> str:
    ip = forwarded.split(",")[0] if (forwarded := request.headers.get("X-Forwarded-For")) else request.client.host
    return ip + ":" + request.scope["path"]


async def http_default_callback(request: Request, response: Response, pexpire: int):
    """
    default callback when too many requests
    :param request:
    :param pexpire: The remaining milliseconds
    :param response:
    :return:
    """
    raise HTTPException(HTTP_429_TOO_MANY_REQUESTS, "Too Many Requests", headers={"Retry-After": str(pexpire)})


async def ws_default_callback(ws: WebSocket, pexpire: int):
    """
    default callback when too many requests
    :param ws:
    :param pexpire: The remaining milliseconds
    :return:
    """
    expire = ceil(pexpire / 1000)
    raise HTTPException(HTTP_429_TOO_MANY_REQUESTS, "Too Many Requests", headers={"Retry-After": str(expire)})


class ACTRatelimit:
    __slots__ = ()

    backend: BaseBackend | None = None
    prefix: str = "act-ratelimit"
    lua_sha: str | None = None
    identifier: IDENTIFIER_SIGNATURE = default_identifier
    http_callback: HTTP_CALLBACK_SIGNATURE = http_default_callback
    ws_callback: WS_CALLBACK_SIGNATURE = ws_default_callback
    strategy: RateLimitStrategy = RateLimitStrategy.FIXED_WINDOW

    @classmethod
    async def init(
        cls,
        backend: BaseBackend,
        *,
        prefix: str | None = None,
        identifier: IDENTIFIER_SIGNATURE | None = None,
        http_callback: HTTP_CALLBACK_SIGNATURE | None = None,
        ws_callback: WS_CALLBACK_SIGNATURE | None = None,
        strategy: RateLimitStrategy | None = None,
    ) -> None:
        cls.backend = backend
        if prefix:
            cls.prefix = prefix
        if identifier:
            cls.identifier = identifier
        if http_callback:
            cls.http_callback = http_callback
        if ws_callback:
            cls.ws_callback = ws_callback
        if strategy:
            cls.strategy = strategy

    @classmethod
    async def close(cls) -> None:
        await cls.backend.close()



__version__ = "0.0.1a2"


# Copyright (c) 2025 syncblaze
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import abc

import typing_extensions

from act_ratelimit.constants import RateLimitStrategy

if typing_extensions.TYPE_CHECKING:
    from redis import asyncio as aioredis
    from valkey.asyncio import Valkey


__all__: tuple[str, ...] = ("BaseBackend", "RedisBackend", "ValkeyBackend")


class BaseBackend(abc.ABC):
    @abc.abstractmethod
    async def check(self, key: str, times: int, limit: int, strategy: RateLimitStrategy) -> int:
        """Check if a key has hit the rate limit.

        This method should return the time in milliseconds until the rate limit resets.
        If the rate limit has not been hit, it should return 0.

        Args:
            key: The key to check.
            times: The number of times the key has to be hit to trigger the rate limit.
            limit: How lo
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def close(self) -> None:
        """Close the connection to the backend."""
        raise NotImplementedError


class ValkeyBackend(BaseBackend):
    LUA_SCRIPT: str = """
        local key = KEYS[1]
        local limit = tonumber(ARGV[1])
        local expire_time = ARGV[2]

        local current = tonumber(server.call('get', key) or "0")
        if current > 0 then
            if current + 1 > limit then
                return server.call("PTTL",key)
            else
                server.call("INCR", key)
                return 0
            end
        else
            server.call("SET", key, 1,"px",expire_time)
            return 0
        end
    """

    LUA_SCRIPT_SLIDING_WINDOW: str = """
        local key = KEYS[1]
        local limit = tonumber(ARGV[1])
        local expire_time = tonumber(ARGV[2])
        local current_time = redis.call('TIME')[1]
        local start_time = current_time - expire_time / 1000
        redis.call('ZREMRANGEBYSCORE', key, 0, start_time)
        local current = redis.call('ZCARD', key)

        if current >= limit then
           return redis.call("PTTL",key)
        else
           redis.call("ZADD", key, current_time, current_time)
           redis.call('PEXPIRE', key, expire_time)
           return 0
        end
    """

    LUA_SCRIPT_FIXED_WINDOW_ELASTIIC: str = """
        local key = KEYS[1]
        local limit = tonumber(ARGV[1])
        local expire_time = tonumber(ARGV[2])

        local current = tonumber(server.call('GET', key) or "0")

        if current > 0 then
            if current + 1 > limit then
                local ttl = server.call("PTTL", key)
                if ttl > 0 then
                    server.call("PEXPIRE", key, ttl + expire_time)  -- Extend timeout
                    return ttl + expire_time
                else
                    server.call("SET", key, 1, "px", expire_time)
                    return 0
                end
            else
                server.call("INCR", key)
                return 0
            end
        else
            server.call("SET", key, 1, "px", expire_time)
            return 0
        end
    """

    def __init__(
        self,
        valkey: Valkey,
    ):
        self.valkey: Valkey = valkey
        self.lua_sha_fixed: str | None = None
        self.lua_sha_sliding: str | None = None
        self.lua_sha_fixed_elastic: str | None = None

    async def check(self, key: str, times: int, limit: int, strategy: RateLimitStrategy) -> int:
        if strategy == RateLimitStrategy.FIXED_WINDOW:
            if not self.lua_sha_fixed:
                self.lua_sha_fixed = await self.valkey.script_load(self.LUA_SCRIPT)
            result: str = await self.valkey.evalsha(self.lua_sha_fixed, 1, key, str(times), str(limit))  # pyright: ignore
            return int(result)  # pyright: ignore
        elif strategy == RateLimitStrategy.SLIDING_WINDOW:
            if not self.lua_sha_sliding:
                self.lua_sha_sliding = await self.valkey.script_load(self.LUA_SCRIPT_SLIDING_WINDOW)
            result: str = await self.valkey.evalsha(self.lua_sha_sliding, 1, key, str(times), str(limit))  # pyright: ignore
            return int(result)  # pyright: ignore
        elif strategy == RateLimitStrategy.FIXED_WINDOW_ELASTIC:
            if not self.lua_sha_fixed_elastic:
                self.lua_sha_fixed_elastic = await self.valkey.script_load(self.LUA_SCRIPT_FIXED_WINDOW_ELASTIIC)
            result: str = await self.valkey.evalsha(self.lua_sha_fixed_elastic, 1, key, str(times), str(limit))  # pyright: ignore
            return int(result)  # pyright: ignore

    async def close(self) -> None:
        await self.valkey.aclose()


class RedisBackend(BaseBackend):
    LUA_SCRIPT_FIXED_WINDOW: str = """
        local key = KEYS[1]
        local limit = tonumber(ARGV[1])
        local expire_time = ARGV[2]

        local current = tonumber(redis.call('get', key) or "0")
        if current > 0 then
            if current + 1 > limit then
                return redis.call("PTTL",key)
            else
                redis.call("INCR", key)
                return 0
            end
        else
            redis.call("SET", key, 1,"px",expire_time)
            return 0
        end
    """

    LUA_SCRIPT_SLIDING_WINDOW: str = """
        local key = KEYS[1]
        local limit = tonumber(ARGV[1])
        local expire_time = tonumber(ARGV[2])
        local current_time = redis.call('TIME')[1]
        local start_time = current_time - expire_time / 1000
        redis.call('ZREMRANGEBYSCORE', key, 0, start_time)
        local current = redis.call('ZCARD', key)

        if current >= limit then
            return redis.call("PTTL",key)
        else
            redis.call("ZADD", key, current_time, current_time)
            redis.call('PEXPIRE', key, expire_time)
            return 0
        end
    """

    LUA_SCRIPT_FIXED_WINDOW_ELASTIIC: str = """
    local key = KEYS[1]
    local limit = tonumber(ARGV[1])
    local expire_time = tonumber(ARGV[2])

    local current = tonumber(redis.call('GET', key) or "0")

    if current > 0 then
        if current + 1 > limit then
            local ttl = redis.call("PTTL", key)
            if ttl > 0 then
                redis.call("PEXPIRE", key, ttl + expire_time)  -- Extend timeout
                return ttl + expire_time
            else
                redis.call("SET", key, 1, "px", expire_time)
                return 0
            end
        else
            redis.call("INCR", key)
            return 0
        end
    else
        redis.call("SET", key, 1, "px", expire_time)
        return 0
    end
    """

    def __init__(
        self,
        redis: aioredis.Redis[bytes],
    ):
        self.redis: aioredis.Redis[bytes] = redis
        self.lua_sha_fixed: str | None = None
        self.lua_sha_sliding: str | None = None
        self.lua_sha_fixed_elastic: str | None = None

    async def check(self, key: str, times: int, limit: int, strategy: RateLimitStrategy) -> int:
        if strategy == RateLimitStrategy.FIXED_WINDOW:
            if not self.lua_sha_fixed:
                self.lua_sha_fixed = await self.redis.script_load(self.LUA_SCRIPT_FIXED_WINDOW)  # pyright: ignore
            result: str = await self.redis.evalsha(self.lua_sha_fixed, 1, key, str(times), str(limit))  # pyright: ignore
            return int(result)  # pyright: ignore
        elif strategy == RateLimitStrategy.SLIDING_WINDOW:
            if not self.lua_sha_sliding:
                self.lua_sha_sliding = await self.redis.script_load(self.LUA_SCRIPT_SLIDING_WINDOW)  # pyright: ignore
            result: str = await self.redis.evalsha(self.lua_sha_sliding, 1, key, str(times), str(limit))  # pyright: ignore
            return int(result)  # pyright: ignore
        elif strategy == RateLimitStrategy.FIXED_WINDOW_ELASTIC:
            if not self.lua_sha_fixed_elastic:
                self.lua_sha_fixed_elastic = await self.redis.script_load(self.LUA_SCRIPT_FIXED_WINDOW_ELASTIIC)  # pyright: ignore
            result: str = await self.redis.evalsha(self.lua_sha_fixed_elastic, 1, key, str(times), str(limit))  # pyright: ignore
            return int(result)  # pyright: ignore

    async def close(self) -> None:
        await self.redis.aclose()  # pyright: ignore



# -*- coding: utf-8 -*-
# Copyright (c) 2023-present tandemdude
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import argparse
import re
import sys

VERSION_REGEX = re.compile(r"__version__\s*=\s*\"(?P<version>\d+\.\d+\.\d+)(?:a(?P<alphanum>\d+))?\"")

parser = argparse.ArgumentParser()
parser.add_argument("type", choices=["major", "minor", "patch", "alpha"])
parser.add_argument("increment")

_noop = lambda n: n  # noqa: E731
_increment = lambda n: str(int(n) + 1)  # noqa: E731
_reset = lambda _: "0"  # noqa: E731

ACTIONS = {
    "major": (_increment, _reset, _reset, _reset),
    "minor": (_noop, _increment, _reset, _reset),
    "patch": (_noop, _noop, _increment, _reset),
    "alpha": (_noop, _noop, _noop, _increment),
}


def run(version_type: str, increment: str) -> None:
    with open("act_ratelimit/__init__.py") as fp:
        content = fp.read()

    current_version = VERSION_REGEX.search(content)
    if current_version is None:
        raise RuntimeError("Could not find version in __init__ file")

    version = current_version.groupdict()["version"]
    alpha = current_version.groupdict().get("alphanum", "0")

    if increment.lower() != "true":
        sys.stdout.write(f"{version}{('a' + alpha) if version_type.lower() == 'alpha' else ''}")
        return

    major, minor, patch = version.split(".")

    actions = ACTIONS[version_type.lower()]
    major, minor, patch, alpha = [a(s) for a, s in zip(actions, [major, minor, patch, alpha])]

    new_version = f"{major}.{minor}.{patch}"

    if version_type.lower() == "alpha":
        new_version += f"a{alpha}"

    updated_file_content = VERSION_REGEX.sub(f'__version__ = "{new_version}"', content)
    with open("act_ratelimit/__init__.py", "w") as fp:
        fp.write(updated_file_content)

    sys.stdout.write(new_version)


if __name__ == "__main__":
    args = parser.parse_args()
    run(args.type, args.increment)


# Copyright (c) 2025 syncblaze
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from fastapi.routing import APIRoute
from starlette.requests import Request
from starlette.responses import Response
from starlette.websockets import WebSocket

from act_ratelimit import ACTRatelimit
from act_ratelimit.constants import HTTP_CALLBACK_SIGNATURE
from act_ratelimit.constants import IDENTIFIER_SIGNATURE
from act_ratelimit.constants import WS_CALLBACK_SIGNATURE
from act_ratelimit.constants import RateLimitStrategy

__all__: tuple[str, ...] = ("RateLimiter", "WebSocketRateLimiter")


class RateLimiter:
    def __init__(
        self,
        *,
        times: int = 1,
        milliseconds: int = 0,
        seconds: int = 0,
        minutes: int = 0,
        hours: int = 0,
        identifier: IDENTIFIER_SIGNATURE | None = None,
        callback: HTTP_CALLBACK_SIGNATURE | None = None,
        strategy: RateLimitStrategy = RateLimitStrategy.FIXED_WINDOW,
    ):
        self.times = times
        assert self.times >= 0, "times must be greater than or equal to 0"
        self.milliseconds = milliseconds + 1000 * seconds + 60000 * minutes + 3600000 * hours
        assert self.milliseconds >= 0, "time must be greater than or equal to 0"
        self.identifier = identifier
        self.callback = callback
        self.strategy = strategy

        self.route_index = 0
        self.dep_index = 0

    def _set_indexes(self, request: Request):
        for i, route in enumerate(request.app.routes):
            if not isinstance(route, APIRoute):
                continue
            if route.path == request.scope["path"] and request.method in route.methods:
                self.route_index = i
                for j, dependency in enumerate(route.dependencies):
                    if self is dependency.dependency:  # type: ignore
                        self.dep_index = j
                        break

    async def __call__(self, request: Request, response: Response):
        assert ACTRatelimit.backend is not None, "You must call FastAPILimiter.init in startup event of fastapi!"
        if self.milliseconds == 0:
            return

        self._set_indexes(request)

        # moved here because constructor run before app startup
        identifier = self.identifier or ACTRatelimit.identifier
        strategy = self.strategy or ACTRatelimit.strategy
        rate_key = await identifier(request)
        key = f"{ACTRatelimit.prefix}:{rate_key}:{self.route_index}:{self.dep_index}"
        print(key)
        pexpire = await ACTRatelimit.backend.check(key, self.times, self.milliseconds, strategy)
        if pexpire != 0:
            callback = self.callback or ACTRatelimit.http_callback
            return await callback(request, response, pexpire)


class WebSocketRateLimiter:
    def __init__(
        self,
        *,
        times: int = 1,
        milliseconds: int = 0,
        seconds: int = 0,
        minutes: int = 0,
        hours: int = 0,
        identifier: IDENTIFIER_SIGNATURE | None = None,
        callback: WS_CALLBACK_SIGNATURE | None = None,
        strategy: RateLimitStrategy = RateLimitStrategy.FIXED_WINDOW,
    ):
        self.times = times
        assert self.times >= 0, "times must be greater than or equal to 0"
        self.milliseconds = milliseconds + 1000 * seconds + 60000 * minutes + 3600000 * hours
        assert self.milliseconds >= 0, "time must be greater than or equal to 0"
        self.identifier = identifier
        self.callback = callback
        self.strategy = strategy

    async def __call__(self, ws: WebSocket, context_key: str = ""):
        assert ACTRatelimit.backend is not None, "You must call FastAPILimiter.init in startup event of fastapi!"
        if self.milliseconds == 0:
            return
        identifier = self.identifier or ACTRatelimit.identifier
        strategy = self.strategy or ACTRatelimit.strategy
        rate_key = await identifier(ws)
        key = f"{ACTRatelimit.prefix}:ws:{rate_key}:{context_key}"
        pexpire = await ACTRatelimit.backend.check(key, self.times, self.milliseconds, strategy)
        if pexpire != 0:
            callback = self.callback or ACTRatelimit.ws_callback
            return await callback(ws, pexpire)



from time import sleep

from starlette.testclient import TestClient

from examples.main import app


def test_limiter():
    with TestClient(app) as client:
        response = client.get("/")
        assert response.status_code == 200

        response = client.get("/")
        assert response.status_code == 200

        response = client.get("/")
        assert response.status_code == 429

        response = client.post("/")
        assert response.status_code == 200

        response = client.post("/")
        assert response.status_code == 429
        sleep(5)

        response = client.get("/")
        assert response.status_code == 200

        response = client.post("/")
        assert response.status_code == 200


def test_limiter_multiple():
    with TestClient(app) as client:
        response = client.get("/multiple")
        assert response.status_code == 200

        response = client.get("/multiple")
        assert response.status_code == 429
        sleep(5)

        response = client.get("/multiple")
        assert response.status_code == 200

        response = client.get("/multiple")
        assert response.status_code == 429
        sleep(10)

        response = client.get("/multiple")
        assert response.status_code == 200


def test_limiter_websockets():
    with TestClient(app) as client:
        with client.websocket_connect("/ws") as ws:
            ws.send_text("Hi")
            data = ws.receive_text()
            assert data == "Hello, world"

            ws.send_text("Hi")
            data = ws.receive_text()
            assert data == "Hello again"

            ws.send_text("Hi 2")
            data = ws.receive_text()
            assert data == "Hello, world"
            ws.close()



# Copyright (c) 2025 syncblaze
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import typing as t
from enum import IntEnum

from starlette.requests import Request
from starlette.responses import Response
from starlette.websockets import WebSocket

__all__: tuple[str, ...] = (
    "HTTP_CALLBACK_SIGNATURE",
    "IDENTIFIER_SIGNATURE",
    "WS_CALLBACK_SIGNATURE",
    "RateLimitStrategy",
)


class RateLimitStrategy(IntEnum):
    """
    Rate limit strategies.
    """

    FIXED_WINDOW = 0
    SLIDING_WINDOW = 1
    FIXED_WINDOW_ELASTIC = 2


HTTP_CALLBACK_SIGNATURE = t.Callable[[Request, Response, int], t.Awaitable[t.Any]]
IDENTIFIER_SIGNATURE = t.Callable[[Request | WebSocket], t.Awaitable[str]]
WS_CALLBACK_SIGNATURE = t.Callable[[WebSocket, int], t.Awaitable[t.Any]]



# -*- coding: utf-8 -*-
# Copyright (c) 2023-present tandemdude
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import typing as t
from collections.abc import Callable

import nox
from nox import options

SCRIPT_PATHS = [
    os.path.join(".", "examples"),
    os.path.join(".", "act_ratelimit"),
    "noxfile.py",
]

options.sessions = ["format_fix", "typecheck"]


def nox_session(**kwargs: t.Any) -> Callable[[Callable[[nox.Session], None]], Callable[[nox.Session], None]]:
    kwargs.setdefault("venv_backend", "uv|virtualenv")
    kwargs.setdefault("reuse_venv", True)

    def inner(func: Callable[[nox.Session], None]) -> Callable[[nox.Session], None]:
        return nox.session(**kwargs)(func)

    return inner


@nox_session()
def format_fix(session: nox.Session) -> None:
    session.install("-U", ".[valkey,redis,dev.format]")
    session.run("python", "-m", "ruff", "format", *SCRIPT_PATHS)
    session.run("python", "-m", "ruff", "check", "--fix", *SCRIPT_PATHS)


@nox_session()
def format_check(session: nox.Session) -> None:
    session.install("-U", ".[valkey,redis,dev.format]")
    session.run("python", "-m", "ruff", "format", *SCRIPT_PATHS, "--check")
    session.run("python", "-m", "ruff", "check", "--output-format", "github", *SCRIPT_PATHS)


@nox_session()
def typecheck(session: nox.Session) -> None:
    session.install("-U", ".[valkey,redis,dev.typecheck]")
    session.run("python", "-m", "pyright")


@nox_session()
def test(session: nox.Session) -> None:
    session.install("-U", ".[localization,crontrigger,dev.test]")

    args = ["python", "-m", "pytest"]
    if session.posargs:
        args.extend(["--cov", "lightbulb"])
    args.append("tests")

    session.run(*args)



from contextlib import asynccontextmanager

import redis.asyncio as redis
import uvicorn
from fastapi import Depends
from fastapi import FastAPI
from fastapi import HTTPException
from fastapi import WebSocket
from fastapi_ratelimit import FastAPILimiter
from fastapi_ratelimit.depends import RateLimiter
from fastapi_ratelimit.depends import WebSocketRateLimiter


@asynccontextmanager
async def lifespan(_: FastAPI):
    redis_connection = redis.from_url("redis://localhost:6379", encoding="utf8")
    await FastAPILimiter.init(redis_connection)
    yield
    await FastAPILimiter.close()


app = FastAPI(lifespan=lifespan)


@app.get("/", dependencies=[Depends(RateLimiter(times=2, seconds=5))])
async def index_get():
    return {"msg": "Hello World"}


@app.post("/", dependencies=[Depends(RateLimiter(times=1, seconds=5))])
async def index_post():
    return {"msg": "Hello World"}


@app.get(
    "/multiple",
    dependencies=[
        Depends(RateLimiter(times=1, seconds=5)),
        Depends(RateLimiter(times=2, seconds=15)),
    ],
)
async def multiple():
    return {"msg": "Hello World"}


@app.websocket("/ws")
async def websocket_endpoint(websocket: WebSocket):
    await websocket.accept()
    ratelimit = WebSocketRateLimiter(times=1, seconds=5)
    while True:
        try:
            data = await websocket.receive_text()
            await ratelimit(websocket, context_key=data)  # NB: context_key is optional
            await websocket.send_text("Hello, world")
        except HTTPException:
            await websocket.send_text("Hello again")


if __name__ == "__main__":
    uvicorn.run("main:app", debug=True, reload=True)
